
package com.sample;

import java.util.Map;
import java.util.WeakHashMap;

import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.os.IBinder;
import android.widget.Toast;

import com.felicanetworks.mfc.AppInfo;
import com.felicanetworks.mfc.Felica;
import com.felicanetworks.mfc.FelicaEventListener;
import com.felicanetworks.mfc.FelicaException;
import com.felicanetworks.mfc.PushSegment;

public class FelicaPushSender {
    private Context context;

    // 接続状態
    private boolean connected = false;
    private Felica felica = null;
    private volatile boolean canceled = false;

    public FelicaPushSender(Context context) {
        this.context = context;
    }

    public void connect() throws Exception {

        if (context == null) {
            throw new Exception("connect error:Context is not set.");
        }

        if (connected) {
            // 接続済み
            return;
        }

        Intent intent = new Intent();
        intent.setClass(context, Felica.class);
        if (!context.bindService(intent, instance, Context.BIND_AUTO_CREATE)) {
            throw new Exception("connect error:Context#bindService() failed.");
        }
        // 接続状態変更はonServiceConnected()が呼び出されたタイミングで実施
    }

    public void disconnect() throws Exception {
        if (context == null) {
            throw new Exception("connect error:Context is not set.");
        }

        if (!connected) {
            return;
        }

        context.unbindService(instance);

        // 接続状態を解除
        connected = false;
    }

    public void push(PushSegment pushSegment, FelicaPushListener listener) {

        try {
            felica.open();

            int retryCount = 0;
            canceled = false;
            while (retryCount < 10) { // 10回リトライ
                try {

                    if (canceled) { // ダイアログのキャンセルが押された場合ここで終了する
                        break;
                    }

                    felica.push(pushSegment);
                    break;

                } catch (IllegalArgumentException e) {
                    // Intentの値が大きすぎるなど、不正な引数が指定された場合
                    handleUnexpectedException(e);

                } catch (FelicaException e) {
                    retryCount++;

                    if (FelicaException.ID_IO_ERROR == e.getID()
                            && FelicaException.TYPE_TIMEOUT_OCCURRED == e.getType()) {
                        continue;
                    }
                }
            }

            felica.close();
            felica.inactivateFelica();

        } catch (FelicaException fe) {
            handleFelicaException(fe);
        }

        if (canceled) {
            listener.finished(false);
        } else {
            listener.finished(true);
        }
    }

    public void activate(FelicaEventListener listener) {

        if (felica == null) {
            throw new IllegalArgumentException("Felica is not set");
        }

        try {
            felica.activateFelica(null, listener);

        } catch (IllegalArgumentException e) {
            handleUnexpectedException(e);
        } catch (FelicaException e) {
            handleFelicaException(e);
        }

    }

    private final ServiceConnection instance = new ServiceConnection() {

        @Override
        public void onServiceConnected(ComponentName name, IBinder service) {

            // Felicaとの接続が確立されたので、Felicaインスタンスを取得する
            felica = ((Felica.LocalBinder) service).getInstance();
            connected = true;
            canceled = false;

        }

        @Override
        public void onServiceDisconnected(ComponentName name) {

            // Felicaの設定解除
            felica = null;
            connected = false;
        }
    };

    /**
     * FelicaExceptionの種別を文字列に変換するためのマップ。
     */
    private static final Map<Integer, String> FELICA_EXCEPTION_ID_CONVERSION_MAP = new WeakHashMap<Integer, String>() {

        {
            put(FelicaException.ID_UNKNOWN_ERROR, "ID_UNKNOWN_ERROR");
            put(FelicaException.ID_ILLEGAL_STATE_ERROR, "ID_ILLEGAL_STATE_ERROR");
            put(FelicaException.ID_IO_ERROR, "ID_IO_ERROR");
            put(FelicaException.ID_GET_KEY_VERSION_ERROR, "ID_GET_KEY_VERSION_ERROR");
            put(FelicaException.ID_READ_ERROR, "ID_READ_ERROR");
            put(FelicaException.ID_WRITE_ERROR, "ID_WRITE_ERROR");
            put(FelicaException.ID_SET_NODECODESIZE_ERROR, "ID_SET_NODECODESIZE_ERROR");
            put(FelicaException.ID_OPEN_ERROR, "ID_OPEN_ERROR");
            put(FelicaException.ID_GET_NODE_INFORMATION_ERROR, "ID_GET_NODE_INFORMATION_ERROR");
            put(FelicaException.ID_GET_PRIVACY_NODE_INFORMATION_ERROR,
                    "ID_GET_PRIVACY_NODE_INFORMATION_ERROR");
            put(FelicaException.ID_SET_PRIVACY_ERROR, "ID_SET_PRIVACY_ERROR");
            put(FelicaException.ID_PERMISSION_ERROR, "ID_PERMISSION_ERROR");
            put(FelicaException.ID_GET_BLOCK_COUNT_INFORMATION_ERROR,
                    "ID_GET_BLOCK_COUNT_INFORMATION_ERROR");
        }
    };

    /**
     * FelicaExceptionのタイプを文字列に変換するためのマップ。
     */
    private static final Map<Integer, String> FELICA_EXCEPTION_TYPE_CONVERSION_MAP = new WeakHashMap<Integer, String>() {

        {
            put(FelicaException.TYPE_NOT_OPENED, "TYPE_NOT_OPENED");
            put(FelicaException.TYPE_CURRENTLY_ONLINE, "TYPE_CURRENTLY_ONLINE");
            put(FelicaException.TYPE_NOT_SELECTED, "TYPE_NOT_SELECTED");
            put(FelicaException.TYPE_NOT_ACTIVATED, "TYPE_NOT_ACTIVATED");
            put(FelicaException.TYPE_INVALID_RESPONSE, "TYPE_INVALID_RESPONSE");
            put(FelicaException.TYPE_TIMEOUT_OCCURRED, "TYPE_TIMEOUT_OCCURRED");
            put(FelicaException.TYPE_OPEN_FAILED, "TYPE_OPEN_FAILED");
            put(FelicaException.TYPE_SELECT_FAILED, "TYPE_SELECT_FAILED");
            put(FelicaException.TYPE_GET_KEY_VERSION_FAILED, "TYPE_GET_KEY_VERSION_FAILED");
            put(FelicaException.TYPE_SERVICE_NOT_FOUND, "TYPE_SERVICE_NOT_FOUND");
            put(FelicaException.TYPE_BLOCK_NOT_FOUND, "TYPE_BLOCK_NOT_FOUND");
            put(FelicaException.TYPE_PIN_NOT_CHECKED, "TYPE_PIN_NOT_CHECKED");
            put(FelicaException.TYPE_READ_FAILED, "TYPE_READ_FAILED");
            put(FelicaException.TYPE_PURSE_FAILED, "TYPE_PURSE_FAILED");
            put(FelicaException.TYPE_CASH_BACK_FAILED, "TYPE_CASH_BACK_FAILED");
            put(FelicaException.TYPE_INVALID_PIN, "TYPE_INVALID_PIN");
            put(FelicaException.TYPE_CHECK_PIN_LIMIT, "TYPE_CHECK_PIN_LIMIT");
            put(FelicaException.TYPE_CHECK_PIN_OVERRUN, "TYPE_CHECK_PIN_OVERRUN");
            put(FelicaException.TYPE_WRITE_FAILED, "TYPE_WRITE_FAILED");
            put(FelicaException.TYPE_ENABLE_PIN_FAILED, "TYPE_ENABLE_PIN_FAILED");
            put(FelicaException.TYPE_FELICA_NOT_SET, "TYPE_FELICA_NOT_SET");
            put(FelicaException.TYPE_DEVICELIST_NOT_SET, "TYPE_DEVICELIST_NOT_SET");
            put(FelicaException.TYPE_LISTENER_NOT_SET, "TYPE_LISTENER_NOT_SET");
            put(FelicaException.TYPE_COMMUNICATION_START_FAILED, "TYPE_COMMUNICATION_START_FAILED");
            put(FelicaException.TYPE_SET_NODECODESIZE_FAILED, "TYPE_SET_NODECODESIZE_FAILED");
            put(FelicaException.TYPE_GET_CONTAINER_ISSUE_INFORMATION_FAILED,
                    "TYPE_GET_CONTAINER_ISSUE_INFORMATION_FAILED");
            put(FelicaException.TYPE_NOT_IC_CHIP_FORMATTING, "TYPE_NOT_IC_CHIP_FORMATTING");
            put(FelicaException.TYPE_ILLEGAL_NODECODE, "TYPE_ILLEGAL_NODECODE");
            put(FelicaException.TYPE_GET_NODE_INFORMATION_FAILED,
                    "TYPE_GET_NODE_INFORMATION_FAILED");
            put(FelicaException.TYPE_GET_PRIVACY_NODE_INFORMATION_FAILED,
                    "TYPE_GET_PRIVACY_NODE_INFORMATION_FAILED");
            put(FelicaException.TYPE_SET_PRIVACY_FAILED, "TYPE_SET_PRIVACY_FAILED");
            put(FelicaException.TYPE_NOT_CLOSED, "TYPE_NOT_CLOSED");
            put(FelicaException.TYPE_ILLEGAL_METHOD_CALL, "TYPE_ILLEGAL_METHOD_CALL");
            put(FelicaException.TYPE_PUSH_FAILED, "TYPE_PUSH_FAILED");
            put(FelicaException.TYPE_ALREADY_ACTIVATED, "TYPE_ALREADY_ACTIVATED");
            put(FelicaException.TYPE_GET_BLOCK_COUNT_INFORMATION_FAILED,
                    "TYPE_GET_BLOCK_COUNT_INFORMATION_FAILED");
            put(FelicaException.TYPE_RESET_FAILED, "TYPE_RESET_FAILED");
            put(FelicaException.TYPE_GET_SYSTEM_CODE_LIST_FAILED,
                    "TYPE_GET_SYSTEM_CODE_LIST_FAILED");
            put(FelicaException.TYPE_GET_CONTAINER_ID_FAILED, "TYPE_GET_CONTAINER_ID_FAILED");
            put(FelicaException.TYPE_REMOTE_ACCESS_FAILED, "TYPE_REMOTE_ACCESS_FAILED");
            put(FelicaException.TYPE_CURRENTLY_ACTIVATING, "TYPE_CURRENTLY_ACTIVATING");
            put(FelicaException.TYPE_ILLEGAL_SYSTEMCODE, "TYPE_ILLEGAL_SYSTEMCODE");
            put(FelicaException.TYPE_GET_RFS_STATE_FAILED, "TYPE_GET_RFS_STATE_FAILED");
            put(FelicaException.TYPE_INVALID_SELECTED_INTERFACE, "TYPE_INVALID_SELECTED_INTERFACE");
            put(FelicaException.TYPE_FELICA_NOT_AVAILABLE, "TYPE_FELICA_NOT_AVAILABLE");
        }
    };

    /**
     * FelicaExceptionの内容に応じて、エラー処理を実施します。 処理の内容は以下の通り。
     * ・ID_UNKNOWN_ERRORが発生した場合に、FeliCaチップのクローズ、およびFeliCaチップの利用終了処理を実施
     * ・(ID_OPEN_ERROR, TYPE_NOT_IC_CHIP_FORMATTING)が発生した場合に、FeliCaチップの利用終了処理を実施
     * ・エラー内容を画面に表示
     * 
     * @param e 処理対象のFelicaException
     */
    private void handleFelicaException(FelicaException e) {

        String errMsg;
        String iDString = FELICA_EXCEPTION_ID_CONVERSION_MAP.get(e.getID());
        String typeString = FELICA_EXCEPTION_TYPE_CONVERSION_MAP.get(e.getType());
        String additionalMsg = null;

        switch (e.getID()) {
            case FelicaException.ID_UNKNOWN_ERROR:
                // 未知のエラーが発生
                // 復帰不能なのでFelica#close()、#inactivateFelica()を実行
                try {
                    felica.close();
                    felica.inactivateFelica();
                } catch (FelicaException e1) {
                    // 強制終了処理に失敗した場合、キャッチした例外を無視
                }
                break;
            case FelicaException.ID_OPEN_ERROR:
                if (e.getType() == FelicaException.TYPE_NOT_IC_CHIP_FORMATTING) {
                    // Felica#inactivateFelica()を実行し、FeliCaチップの利用終了処理を実施
                    try {
                        felica.inactivateFelica();
                    } catch (FelicaException e1) {
                        // 利用終了処理に失敗した場合、キャッチした例外を無視
                    }
                }
                break;
            default:
                break;
        }

        // エラー内容をViewに表示

        // ベースメッセージ生成
        errMsg = "caught FelicaException\n(ID, TYPE)=(" + iDString + ", " + typeString + ")";

        // 追加メッセージ作成
        switch (e.getID()) {
            case FelicaException.ID_UNKNOWN_ERROR:
                // 復帰困難なエラーが発生した旨を表示
                additionalMsg = "(Non-recoverable error)";
                break;
            case FelicaException.ID_READ_ERROR:
            case FelicaException.ID_WRITE_ERROR:
            case FelicaException.ID_GET_NODE_INFORMATION_ERROR:
            case FelicaException.ID_GET_PRIVACY_NODE_INFORMATION_ERROR:
            case FelicaException.ID_GET_BLOCK_COUNT_INFORMATION_ERROR:
            case FelicaException.ID_SET_PRIVACY_ERROR:
                // ステータスフラグ取得
                additionalMsg = "Status Flag1:" + e.getStatusFlag1() + ", StatusFlag2:"
                        + e.getStatusFlag2();
                break;
            default:
                break;
        }

        if (additionalMsg != null) {
            errMsg += "\n";
            errMsg += additionalMsg;
        }

        Toast.makeText(context, errMsg, Toast.LENGTH_SHORT).show();
    }

    /**
     * 予期しないエラーへの処理を実施します(予期しないエラーか否かは呼出元で判定済)。 処理の内容は以下の通り。
     * ・FeliCaチップのクローズ、およびFeliCaチップの利用終了処理を実施 ・エラー内容を画面に表示
     * 
     * @param e 処理対象のException
     */
    private void handleUnexpectedException(Exception e) {

        // 復帰不能なのでFelica#close()、#inactivateFelica()を実行
        try {
            felica.close();
            felica.inactivateFelica();
        } catch (FelicaException e1) {
            // 強制終了処理に失敗した場合、キャッチした例外を無視
        }

    }

    public void errorOccurred(int id, String msg, AppInfo otherAppInfo) {

        String result = "Felica#activateFelica() failed\nError:";

        switch (id) {
            case FelicaEventListener.TYPE_USED_BY_OTHER_APP:
                // 他アプリケーションによってFeliCaチップが使用中の場合
                result += "FeliCa Chip is used by other application(PID:" + otherAppInfo.getPid()
                        + ")";
                break;
            case FelicaEventListener.TYPE_NOT_FOUND_ERROR:
                // 利用可能な許可証がみつからなかった場合
                result += "valid permits not found";
                break;
            case FelicaEventListener.TYPE_HTTP_ERROR:
                // 許可証検証中にHTTPエラーが発生した場合
                result += "HTTP error";
                break;
            case FelicaEventListener.TYPE_MFC_VERSION_ERROR:
                // MFCのバージョンアップが必要な場合
                result += "MFC version is too old";
                break;
            case FelicaEventListener.TYPE_UTILITY_VERSION_ERROR:
                // MFCユーティリティのバージョンアップが必要な場合
                result += "MFC utility version is too old";
                break;
            case FelicaEventListener.TYPE_UNKNOWN_ERROR:
                // その他のエラー
                result += "unknown error";
                break;
            default:
                // MFC仕様上、想定外のパス
                result += "unexpected error";
                break;
        }
        if (msg != null) {
            result += "\nerror detail:" + msg;
        }

        Toast.makeText(context, result, Toast.LENGTH_LONG).show();
    }

    public void canceled() {
        canceled = true;
    }

}
